/*
Copyright 2015-2022 Ian Jauslin

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

#include "tree.h"
#include <stdlib.h>
#include <stdio.h>
#include "array.h"

// init
int init_Tree(Tree* tree, int memory_children, int memory_label){
  init_Char_Array(&(tree->root_label),memory_label);
  (*tree).children=calloc(memory_children,sizeof(Tree));
  (*tree).memory=memory_children;
  (*tree).length=0;
  return(0);
}
int free_Tree(Tree tree){
  int i;
  free_Char_Array(tree.root_label);
  for(i=0;i<tree.length;i++){
    free_Tree(tree.children[i]);
  }
  free(tree.children);
  return(0);
}

// copy
int tree_cpy(Tree input, Tree* output){
  init_Tree(output,input.length, input.root_label.length);
  tree_cpy_noinit(input,output);
  return(0);
}
int tree_cpy_noinit(Tree input, Tree* output){
  int i;
  if((*output).memory<input.length){
    fprintf(stderr,"error: trying to copy a tree of length %d to another with memory %d\n",input.length, (*output).memory);
    exit(-1);
  }
  char_array_cpy_noinit(input.root_label,&(output->root_label));
  for(i=0;i<input.length;i++){
    tree_cpy(input.children[i],(*output).children+i);
  }

  (*output).length=input.length;
  return(0);
}

// resize memory
int tree_resize(Tree* tree, int newsize){
  Tree new_tree;
  init_Tree(&new_tree,newsize,tree->root_label.memory);
  tree_cpy_noinit(*tree,&new_tree);
  free_Tree(*tree);
  *tree=new_tree;
  return(0);
}

// set label
int tree_set_label(Char_Array label, Tree* tree){
  if(label.length > tree->root_label.memory){
    char_array_resize(&(tree->root_label),label.length);
  }
  char_array_cpy_noinit(label,&(tree->root_label));
  return(0);
}

// add a child to a tree
int tree_append_child(Tree child, Tree* output){
  if((*output).length>=(*output).memory){
    tree_resize(output,2*(*output).memory+1);
  }
  tree_cpy(child,(*output).children+(*output).length);
  (*output).length++;
  return(0);
}
// add a child to a tree without allocating memory for the new child
int tree_append_child_noinit(Tree child, Tree* output){
  if((*output).length>=(*output).memory){
    tree_resize(output,2*(*output).memory+1);
  }
  (*output).children[(*output).length]=child;
  (*output).length++;
  return(0);
}

// concatenate the children of two trees
int tree_concat_children(Tree input, Tree* output){
  int i;
  for(i=0;i<input.length;i++){
    tree_append_child(input.children[i],output);
  }
  return(0);
}
// noinit
int tree_concat_children_noinit(Tree input, Tree* output){
  int i;
  for(i=0;i<input.length;i++){
    tree_append_child_noinit(input.children[i],output);
  }
  // free input array
  free(input.children);
  return(0);
}

